# Code written by Theodore Lindsey for his master's thesis.
# Fall 2016
# Liscensed under the GNU Lesser General Public License v3
from LERSdat import LERSdat
from Rule import Rule
import itertools
from utilities import union, intersection, isnum, progressBar, nCr

class IRIM:
    def __init__(self, filename=None):
        self.rules = None
        self.lers = LERSdat()
        if filename is not None:
            self.open(filename)


    def open(self, filename):
        self.rules = None
        self.lers.open(filename)

    def computeAllRules(self, **options):
        # establish option parameters
        minLen = options.get("minLen")
        if minLen is None:
            minLen = 2
        maxLen = options.get("maxLen")
        if maxLen is None:
            maxLen = minLen
        ratio = options.get("ratio")
        if ratio is None:
            ratio = 0.8
        minStr = options.get("minStr")
        if minStr is None:
            minStr = 1
        minCov = options.get("minCov")
        if minCov is None:
            minCov = 1

        # don't let the max length get out of hand
        maxLen = min(maxLen, len(self.lers.attributes))

        # find all possible conditions (rules)
        possibleRules = []
        self.rules = []
        for length in range(minLen, maxLen+1):
            print("computing possible rules of length {}".format(length))
            count = nCr(len(self.lers.attributes), length)
            for idx, attributeSet in enumerate(itertools.combinations(self.lers.attributes, length)):
                print(progressBar((idx+1) / count), end="\r")
                #valueSet = [[val for val in self.blocks[attr].keys() if (len(self.blocks[attr][val])/len(self.lers.records) > ratio)] for attr in attributeSet]
                valueSet = [[val for val in self.lers.blocks[attr].keys() if val not in ["*", "?", "-"]] for attr in attributeSet]

                for combination in itertools.product(*valueSet):
                    possibleRules.append(Rule(list(zip(attributeSet, combination))))

        print("\n\nThere are a total of {} candidate rules".format(len(possibleRules)))

        # trim the list down to just actual rules
        concepts = self.lers.index[self.lers.decision]
        print("\nremoving out-of-spec rules")
        for idx, rule in enumerate(possibleRules):
            #print()
            print(progressBar((idx+1) / len(possibleRules)), end="\r")

            coverage = intersection([self.lers.blocks[attr][val] for attr,val in rule.lhs])

            if len(coverage) < minCov and len(coverage) < minStr:
                continue

            coverageRatios = self.coverageRatioCheck(coverage)
            bestDecision = max(coverageRatios, key=lambda item: item[1])

            concept = self.lers.index[self.lers.decision][bestDecision[0]]
            rule.correct = bestDecision[2]

            if bestDecision[1] > ratio and rule.strength() >= minStr:
                rule.incorrect = coverage.difference(bestDecision[2])
                rule.rhs = (self.lers.decision, bestDecision[0])
                self.rules.append(rule)

        # sort the rules by strength
        self.rules.sort(key=lambda item: item.strength(), reverse=True)

        print()

        return self.rules

    def coverageRatioCheck(self, covered):
        coverageRatios = []

        #covered = intersection([self.lers.blocks[attr][val] for attr,val in lhs])

        for d in self.decisionValues():
            correct = intersection([covered, self.lers.blocks[self.lers.decision][d]])
            if len(covered) > 0:
                coverageRatios.append((d, len(correct) / len(covered), correct))
            else:
                coverageRatios.append((d, 0, set()))

        return coverageRatios

    def decisionValues(self):
        return [v for v in self.lers.index[self.lers.decision]]

    def evaluateRules(self, top=None, rules=None):
        if rules is None:
            if self.rules is not None:
                rules = self.rules
            else:
                return

        temp = []
        if top is None or top > len(rules)+1:
            top = len(rules)+1

        for r in rules[:top]:
            lhs, rhs = r.lhs, r.rhs
            lhsConcept = r.coverage

            rhsConcept = self.lers.blocks[rhs[0]][rhs[1]]

            stats = r.stats(rhsConcept)

            for s in stats:
                if s is set and len(s) > 10:
                    s = "too large to display"

            print(r)
            print("\tCoverage         ({}): {}".format(stats["cov#"], stats["cov"]))
            print("\tCorrect/Strength ({}): {}".format(stats["cor#"], stats["cor"]))
            print("\tIncorrect        ({}): {}".format(stats["icor#"], stats["icor"] if stats["icor#"] > 0 else "{}"))
            print("\tP(rhs | lhs) : {}".format(stats["pr"]))
            print("\tSpecificity  : {}".format(stats["sp"]))
            print("\tScore        : {}".format(stats["sc"]))

    def __str__(self):
        if self.rules is None:
            return "No rules computed"
        else:
            s = ""

            for r in self.rules:
                s += "{}\n".format(r)

            return s

    def __getitem__(self, key):
        return self.rules[key]

    def __iter__(self):
        return self.rules.__iter__()

if __name__ == "__main__":
    i = IRIM("./Inputs/jerzy1.txt")
    rules = i.computeAllRules(minLen=2, maxLen=2, ratio=0.8, minStr=1)
    i.evaluateRules(rules)

