# Code written by Theodore Lindsey for his master's thesis.
# Fall 2016
# Liscensed under the GNU Lesser General Public License v3
import pandas as pd
from utilities import isnum, progressBar, fileLineCount

class LERSdat:
    def __init__(self, filename=None):
        self.attributes = []
        self.attributeTypes = []
        self.records = {}
        self.index = {}
        self.decision = ""
        self.decisions = {}
        self.blocks = {}

        if filename is not None:
            self.open(filename)


    def open(self, filename):
        self.attributes = []
        self.attributeTypes = []
        self.records = {}
        self.index = {}
        self.decision = ""
        self.decisions = {}
        self.blocks = {}

        self.parse(filename)
        self.computeAllBlocks()


    def parse(self, filename):
        # parses the file to extract the dataset
        # file must be in LERS format
        linecount = fileLineCount(filename)
        phase = "find attributes"

        file = open(filename, 'r')
        case = 0
        attribute = 0
        caseValues = []

        print("Loading from file {}".format(filename))
        for linecurrent, line in enumerate(file):
            print(progressBar((linecurrent+1)/linecount), end="\r")

            line = line.split()
            if not line:
                continue

            elif phase == "find attributes" and line[0].startswith("["):
                # this mode finds the attributes
                # attributes should be the second line in the file
                if line[-1].endswith("]"):
                    # if the line also ends the attributes, move to next step
                    self.attributes = line[1:-2]
                    self.decision = line[-2]

                    self.records = pd.DataFrame(columns = self.attributes + [self.decision])

                    # set up voting for numerical attribute types
                    self.attributeTypes = [0 for x in self.attributes]
                    phase = "read records"
                else:
                    # otherwise, keep reading attributes
                    phase = "reading attributes"
                    line = line[1:]

            elif phase == "reading attributes":
                # keep reading attributes
                if line[-1].endswith("]"):
                    # we found the end of the attribute list
                    self.attributes += line[:-2]
                    self.decision = line[-2]

                    self.records = pd.DataFrame(columns = self.attributes + [self.decision])

                    # set up voting for numerical attribute types
                    self.attributeTypes = [0 for x in self.attributes]
                    phase = "read records"
                else:
                    # we're still looking for attributes
                    self.attributes += line

            elif phase == "read records":
                # we finished reading attributes and now we're reading values for cases
                for element in line:
                    # TODO: (optimization) if isnum(element), element = float(element)

                    if element.startswith("!"):
                        # ingore the rest of the line.  it's a comment
                        break


                    if attribute + 1 > len(self.attributes):
                        # we've found all the attributes for this case

                        if self.decision not in self.index:
                            self.index[self.decision] = {}
                        if element not in self.index[self.decision]:
                            self.index[self.decision][element] = []

                        self.index[self.decision][element].append(case)

                        self.decisions[case] = element

                        caseValues.append(element)
                        # vote on if the attribute is a number or not
                        if len(self.attributeTypes) < len(caseValues):
                            self.attributeTypes.append(int(isnum(element)))
                        else:
                            self.attributeTypes[len(caseValues) - 1] += int(isnum(element))

                        self.records.loc[case] = caseValues

                        case += 1
                        attribute = 0
                        caseValues = []
                        continue

                    caseValues.append(element)
                    # vote on if the attribute is a number or not
                    self.attributeTypes[len(caseValues) - 1] += int(isnum(element))

                    if self.attributes[attribute] not in self.index:
                        self.index[self.attributes[attribute]] = {}
                    if element not in self.index[self.attributes[attribute]]:
                        self.index[self.attributes[attribute]][element] = set([])
                    self.index[self.attributes[attribute]][element].add(case)

                    attribute += 1

        # determine if attribute was numerical
        # todo: (optimization) base numerical decision on what values are, not how many can convert to float

        numcases = len(self.records.columns)
        tempHeader = self.attributes + [self.decision]
        tempAttributeTypes = self.attributeTypes
        self.attributeTypes = {}

        for i in range(len(tempAttributeTypes)):
            if len(set(self.records[(self.attributes + [self.decision])[i]])) == 2:
                self.attributeTypes[tempHeader[i]] = "Binary"
            elif tempAttributeTypes[i] > numcases / 2:
                self.attributeTypes[tempHeader[i]] = "Numerical"
            else:
                self.attributeTypes[tempHeader[i]] = "Discrete"
        print("\n")



    def computeAllBlocks(self):
        for attr in (self.attributes + [self.decision]):
            self.blocks[attr] = {}

        print("computing numerical blocks")
        self.computeNumericalBlocks()
        print("computing discrete blocks")
        self.computeDiscreteBlocks()

        # add all the "do-not-care" cases to the blocks for that attribute
        for attr in self.attributes:
            if "*" in self.blocks[attr]:
                for val in self.blocks[attr]:
                    if val not in ["*", "?", "-"]:
                        self.blocks[attr][val] = self.blocks[attr][val].union(self.blocks[attr]["*"])


    def computeNumericalBlocks(self):
        total = len(self.attributes + [self.decision])
        for idx, attribute in enumerate(self.attributes + [self.decision]):
            print(progressBar((idx+1) / total), end="\r")
            if self.attributeTypes[attribute] == "Numerical":
                # find all unique values
                values = sorted(set(self.records[attribute]).difference(set(["*", "?", "-"])), key=lambda item: float(item))

                # compute the cutpoints for this attribute
                cutpoints = []
                for i in range(len(values)-1):
                    cutpoints.append((float(values[i])+float(values[i+1])) / 2)

                # for each cut point, compute the <= and the > blocks
                for cutpoint in cutpoints:
                    self.blocks[attribute][">{}".format(cutpoint)] = set()
                    self.blocks[attribute]["<={}".format(cutpoint)] = set()
                    for case in self.records.index:
                        if isnum(self.records[attribute][case]):
                            if float(self.records[attribute][case]) > cutpoint:
                                self.blocks[attribute][">{}".format(cutpoint)].add(case)
                            else:
                                self.blocks[attribute]["<={}".format(cutpoint)].add(case)

                # find the "do-not-care" block for numerical attributes, too!
                if "*" in self.index[attribute]:
                    self.blocks[attribute]["*"] = set(self.index[attribute]["*"])
        print()

    def computeDiscreteBlocks(self):
        total = len(self.attributes + [self.decision])
        for (idx, attribute) in enumerate(self.attributes + [self.decision]):
            print(progressBar((idx+1) / total), end="\r")
            if self.attributeTypes[attribute] != "Numerical":
                for value in self.index[attribute]:
                    self.blocks[attribute][value] = set(self.index[attribute][value])
        print()

    def __str__(self):
        # this class will display the index when it is printed
        retStr = ""
        retStr += "Decision:   '{}'\n".format(self.decision)
        retStr += "Attributes: {}\n".format(self.attributes)
        for attribute in self.attributes:
            retStr += str(attribute) + "\n"
            for value in self.index[attribute]:
                retStr += "\t{}: {}\n".format(value, self.index[attribute][value])

        retStr += "{}\n".format(self.decision)
        for value in self.index[self.decision]:
            retStr += "\t{}: {}\n".format(value,self.index[self.decision][value])

        return retStr


if __name__ == "__main__":
    o = LERSdat("../Inputs/jerzy1.txt")
    print(o.records)
    print(o.attributes + [o.decision])
    print(o.attributeTypes)
    o.computeAllBlocks()

    for attr in o.attributes + [o.decision]:
        print("{}:{}".format(attr,o.blocks[attr]))
    #print(o)